@@ -29,6 +29,9 @@
  * contrib/pgcrypto/pgp.h
  */
 
+#include "mbuf.h"
+#include "px.h"
+
 enum PGP_S2K_TYPE
 {
 	PGP_S2K_SIMPLE = 0,
@@ -148,7 +151,9 @@ struct PGP_Context
 	 * internal variables
 	 */
 	int			mdc_checked;
-	int			corrupt_prefix;
+	int			corrupt_prefix; /* prefix failed RFC 4880 "quick check" */
+	int			unsupported_compr;		/* has bzip2 compression */
+	int			unexpected_binary;		/* binary data seen in text_mode */
 	int			in_mdc_pkt;
 	int			use_mdcbuf_filter;
 	PX_MD	   *mdc_ctx;
