@@ -1736,29 +1736,6 @@ create_foreignscan_path(PlannerInfo *root, RelOptInfo *rel,
 {
 	ForeignPath *pathnode = makeNode(ForeignPath);
 
-	/*
-	 * Since the path's required_outer should always include all the rel's
-	 * lateral_relids, forcibly add those if necessary.  This is a bit of a
-	 * hack, but up till early 2019 the contrib FDWs failed to ensure that,
-	 * and it's likely that the same error has propagated into many external
-	 * FDWs.  Don't risk modifying the passed-in relid set here.
-	 */
-	if (rel->lateral_relids && !bms_is_subset(rel->lateral_relids,
-											  required_outer))
-		required_outer = bms_union(required_outer, rel->lateral_relids);
-
-	/*
-	 * Although this function is only designed to be used for scans of
-	 * baserels, before v12 postgres_fdw abused it to make paths for join and
-	 * upper rels.  It will work for such cases as long as required_outer is
-	 * empty (otherwise get_baserel_parampathinfo does the wrong thing), which
-	 * fortunately is the expected case for now.
-	 */
-	if (!bms_is_empty(required_outer) &&
-		!(rel->reloptkind == RELOPT_BASEREL ||
-		  rel->reloptkind == RELOPT_OTHER_MEMBER_REL))
-		elog(ERROR, "parameterized foreign joins are not supported yet");
-
 	pathnode->path.pathtype = T_ForeignScan;
 	pathnode->path.parent = rel;
 	pathnode->path.param_info = get_baserel_parampathinfo(root, rel,
@@ -2098,28 +2075,6 @@ reparameterize_path(PlannerInfo *root, Path *path,
 		case T_SubqueryScan:
 			return create_subqueryscan_path(root, rel, path->pathkeys,
 											required_outer);
-		case T_Append:
-			{
-				AppendPath *apath = (AppendPath *) path;
-				List	   *childpaths = NIL;
-				ListCell   *lc;
-
-				/* Reparameterize the children */
-				foreach(lc, apath->subpaths)
-				{
-					Path	   *spath = (Path *) lfirst(lc);
-
-					spath = reparameterize_path(root, spath,
-												required_outer,
-												loop_count);
-					if (spath == NULL)
-						return NULL;
-					childpaths = lappend(childpaths, spath);
-				}
-				return (Path *)
-					create_append_path(rel, childpaths,
-									   required_outer);
-			}
 		default:
 			break;
 	}
