@@ -59,33 +59,6 @@ SELECT '' AS fortyfive, r1.*, r2.*
    WHERE r1.f1 > r2.f1
    ORDER BY r1.f1, r2.f1;
 
--- Test intervals that are large enough to overflow 64 bits in comparisons
-CREATE TEMP TABLE INTERVAL_TBL_OF (f1 interval);
-INSERT INTO INTERVAL_TBL_OF (f1) VALUES
-  ('2147483647 days 2147483647 months'),
-  ('2147483647 days -2147483648 months'),
-  ('1 year'),
-  ('-2147483648 days 2147483647 months'),
-  ('-2147483648 days -2147483648 months');
--- these should fail as out-of-range
-INSERT INTO INTERVAL_TBL_OF (f1) VALUES ('2147483648 days');
-INSERT INTO INTERVAL_TBL_OF (f1) VALUES ('-2147483649 days');
-INSERT INTO INTERVAL_TBL_OF (f1) VALUES ('2147483647 years');
-INSERT INTO INTERVAL_TBL_OF (f1) VALUES ('-2147483648 years');
-
-SELECT r1.*, r2.*
-   FROM INTERVAL_TBL_OF r1, INTERVAL_TBL_OF r2
-   WHERE r1.f1 > r2.f1
-   ORDER BY r1.f1, r2.f1;
-
-CREATE INDEX ON INTERVAL_TBL_OF USING btree (f1);
-SET enable_seqscan TO false;
-EXPLAIN (COSTS OFF)
-SELECT f1 FROM INTERVAL_TBL_OF r1 ORDER BY f1;
-SELECT f1 FROM INTERVAL_TBL_OF r1 ORDER BY f1;
-RESET enable_seqscan;
-
-DROP TABLE INTERVAL_TBL_OF;
 
 -- Test multiplication and division with intervals.
 -- Floating point arithmetic rounding errors can lead to unexpected results,
