@@ -58,8 +58,8 @@ sub Install
 
 		# suppress warning about harmless redeclaration of $config
 		no warnings 'misc';
-		do "./config_default.pl";
-		do "./config.pl" if (-f "config.pl");
+		require "config_default.pl";
+		require "config.pl" if (-f "config.pl");
 	}
 
 	chdir("../../..")    if (-f "../../../configure");
@@ -220,7 +220,7 @@ sub CopySolutionOutput
 	my $conf   = shift;
 	my $target = shift;
 	my $rem =
-	  qr{Project\("\{8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942\}"\) = "([^"]+)"};
+	  qr{Project\("{8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942}"\) = "([^"]+)"};
 
 	my $sln = read_file("pgsql.sln") || croak "Could not open pgsql.sln\n";
 
@@ -342,25 +342,19 @@ sub GenerateTimezoneFiles
 	my $conf   = shift;
 	my $mf     = read_file("src/timezone/Makefile");
 	$mf =~ s{\\\s*[\r\n]+}{}mg;
-
-	$mf =~ /^TZDATAFILES\s*:?=\s*(.*)$/m
-	  || die "Could not find TZDATAFILES line in timezone makefile\n";
+	$mf =~ /^TZDATA\s*:?=\s*(.*)$/m
+	  || die "Could not find TZDATA line in timezone makefile\n";
 	my @tzfiles = split /\s+/, $1;
 
-	$mf =~ /^POSIXRULES\s*:?=\s*(.*)$/m
-	  || die "Could not find POSIXRULES line in timezone makefile\n";
-	my $posixrules = $1;
-	$posixrules =~ s/\s+//g;
-
 	print "Generating timezone files...";
 
-	my @args = ("$conf/zic/zic", '-d', "$target/share/timezone",
-				'-p', "$posixrules");
+	my @args = ("$conf/zic/zic",
+				'-d',
+				"$target/share/timezone");
 	foreach (@tzfiles)
 	{
 		my $tzfile = $_;
-		$tzfile =~ s|\$\(srcdir\)|src/timezone|;
-		push(@args, $tzfile);
+		push(@args, "src/timezone/data/$tzfile")
 	}
 
 	system(@args);
